﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Orienteering {

  [Item("Orienteering Problem", "Represents a single objective Orienteering Problem.")]
  [Creatable("Problems")]
  [StorableClass]
  public class OrienteeringProblem
    : SingleObjectiveHeuristicOptimizationProblem<IOrienteeringEvaluator, IIntegerVectorCreator>,
    IStorableContent {

    public string Filename { get; set; }

    #region Parameter Properties
    public OptionalValueParameter<DoubleMatrix> CoordinatesParameter {
      get { return (OptionalValueParameter<DoubleMatrix>)Parameters["Coordinates"]; }
    }
    public OptionalValueParameter<DistanceMatrix> DistanceMatrixParameter {
      get { return (OptionalValueParameter<DistanceMatrix>)Parameters["DistanceMatrix"]; }
    }
    public ValueParameter<BoolValue> UseDistanceMatrixParameter {
      get { return (ValueParameter<BoolValue>)Parameters["UseDistanceMatrix"]; }
    }

    public ValueParameter<IntValue> StartingPointParameter {
      get { return (ValueParameter<IntValue>)Parameters["StartingPoint"]; }
    }
    public ValueParameter<IntValue> TerminusPointParameter {
      get { return (ValueParameter<IntValue>)Parameters["TerminusPoint"]; }
    }
    public ValueParameter<DoubleValue> MaximumDistanceParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["MaximumDistance"]; }
    }
    public ValueParameter<DoubleArray> ScoresParameter {
      get { return (ValueParameter<DoubleArray>)Parameters["Scores"]; }
    }
    public ValueParameter<DoubleValue> FixedPenaltyParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["FixedPenalty"]; }
    }

    public OptionalValueParameter<IntegerVector> BestKnownSolutionParameter {
      get { return (OptionalValueParameter<IntegerVector>)Parameters["BestKnownSolution"]; }
    }
    #endregion

    #region Properties
    public DoubleMatrix Coordinates {
      get { return CoordinatesParameter.Value; }
      set { CoordinatesParameter.Value = value; }
    }
    public DistanceMatrix DistanceMatrix {
      get { return DistanceMatrixParameter.Value; }
      set { DistanceMatrixParameter.Value = value; }
    }
    public BoolValue UseDistanceMatrix {
      get { return UseDistanceMatrixParameter.Value; }
      set { UseDistanceMatrixParameter.Value = value; }
    }
    public DoubleValue MaximumDistance {
      get { return MaximumDistanceParameter.Value; }
      set { MaximumDistanceParameter.Value = value; }
    }
    public DoubleArray Scores {
      get { return ScoresParameter.Value; }
      set { ScoresParameter.Value = value; }
    }
    public IntegerVector BestKnownSolution {
      get { return BestKnownSolutionParameter.Value; }
      set { BestKnownSolutionParameter.Value = value; }
    }
    #endregion

    [StorableConstructor]
    private OrienteeringProblem(bool deserializing)
      : base(deserializing) {
    }

    private OrienteeringProblem(OrienteeringProblem original, Cloner cloner)
      : base(original, cloner) {
      RegisterEventHandlers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new OrienteeringProblem(this, cloner);
    }


    public OrienteeringProblem()
      : base(new OrienteeringEvaluator(), new UniformRandomIntegerVectorCreator()) {
      Parameters.Add(new OptionalValueParameter<DoubleMatrix>("Coordinates", "The x- and y-Coordinates of the points."));
      Parameters.Add(new OptionalValueParameter<DistanceMatrix>("DistanceMatrix", "The matrix which contains the distances between the points."));
      Parameters.Add(new ValueParameter<BoolValue>("UseDistanceMatrix", "True if the coordinates based evaluators should calculate the distance matrix from the coordinates and use it for evaluation similar to the distance matrix evaluator, otherwise false.", new BoolValue(true)));
      Parameters.Add(new ValueParameter<IntValue>("StartingPoint", "Index of the starting point.", new IntValue(0)));
      Parameters.Add(new ValueParameter<IntValue>("TerminusPoint", "Index of the ending point.", new IntValue(0)));
      Parameters.Add(new ValueParameter<DoubleValue>("MaximumDistance", "The maximum distance constraint for a Orienteering solution."));
      Parameters.Add(new ValueParameter<DoubleArray>("Scores", "The scores of the points.", new DoubleArray()));
      Parameters.Add(new ValueParameter<DoubleValue>("FixedPenalty", "The penalty for each visited vertex."));
      Parameters.Add(new OptionalValueParameter<IntegerVector>("BestKnownSolution", "The best known solution of this Orienteering instance."));

      Maximization.Value = true;
      MaximizationParameter.Hidden = true;

      SolutionCreator.IntegerVectorParameter.ActualName = "OrienteeringSolution";

      InitializeRandomOrienteeringInstance();

      ParameterizeSolutionCreator();
      ParameterizeEvaluator();

      InitializeOperators();
      RegisterEventHandlers();
    }

    #region Events
    protected override void OnSolutionCreatorChanged() {
      base.OnSolutionCreatorChanged();
      SolutionCreator.IntegerVectorParameter.ActualNameChanged += new EventHandler(SolutionCreator_IntegerVectorParameter_ActualNameChanged);
      ParameterizeSolutionCreator();
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
      ParameterizeOperators();
    }
    protected override void OnEvaluatorChanged() {
      base.OnEvaluatorChanged();
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
    }
    private void SolutionCreator_IntegerVectorParameter_ActualNameChanged(object sender, EventArgs e) {
      ParameterizeEvaluator();
      ParameterizeAnalyzer();
      ParameterizeOperators();
    }
    #endregion

    #region Helpers
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      SolutionCreator.IntegerVectorParameter.ActualNameChanged += new EventHandler(SolutionCreator_IntegerVectorParameter_ActualNameChanged);
      // TODO
    }
    private void ParameterizeSolutionCreator() {
      if (SolutionCreator.LengthParameter.Value == null
          || SolutionCreator.LengthParameter.Value.Value != Scores.Length) {
        SolutionCreator.LengthParameter.Value = new IntValue(Scores.Length);
      }
    }
    private void ParameterizeEvaluator() {
      if (Evaluator is OrienteeringEvaluator) {
        var orienteeringEvaluator = (OrienteeringEvaluator)Evaluator;
        // TODO
      }
    }
    private void ParameterizeAnalyzer() {
      // TODO
    }
    private void InitializeOperators() {
      // TODO  
    }
    private void ParameterizeOperators() {
      // TODO
    }
    #endregion

    private void InitializeRandomOrienteeringInstance() {
      // TODO
    }
  }
}